/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.spark.storage

import scala.collection.Map
import scala.collection.mutable

import org.apache.spark.SparkContext
import org.apache.spark.annotation.DeveloperApi
import org.apache.spark.util.Utils

private[spark]
class StorageStatus(
    val blockManagerId: BlockManagerId,
    val maxMem: Long,
    val blocks: mutable.Map[BlockId, BlockStatus] = mutable.Map.empty) {

  def memUsed() = blocks.values.map(_.memSize).reduceOption(_ + _).getOrElse(0L)

  def memUsedByRDD(rddId: Int) =
    rddBlocks.filterKeys(_.rddId == rddId).values.map(_.memSize).reduceOption(_ + _).getOrElse(0L)

  def diskUsed() = blocks.values.map(_.diskSize).reduceOption(_ + _).getOrElse(0L)

  def diskUsedByRDD(rddId: Int) =
    rddBlocks.filterKeys(_.rddId == rddId).values.map(_.diskSize).reduceOption(_ + _).getOrElse(0L)

  def memRemaining : Long = maxMem - memUsed()

  def rddBlocks = blocks.collect { case (rdd: RDDBlockId, status) => (rdd, status) }
}

@DeveloperApi
private[spark]
class RDDInfo(
    val id: Int,
    val name: String,
    val numPartitions: Int,
    val storageLevel: StorageLevel)
  extends Ordered[RDDInfo] {

  var numCachedPartitions = 0
  var memSize = 0L
  var diskSize = 0L
  var tachyonSize = 0L

  override def toString = {
    import Utils.bytesToString
    ("RDD \"%s\" (%d) Storage: %s; CachedPartitions: %d; TotalPartitions: %d; MemorySize: %s;" +
      "TachyonSize: %s; DiskSize: %s").format(
        name, id, storageLevel.toString, numCachedPartitions, numPartitions,
        bytesToString(memSize), bytesToString(tachyonSize), bytesToString(diskSize))
  }

  override def compare(that: RDDInfo) = {
    this.id - that.id
  }
}

/* Helper methods for storage-related objects */
private[spark]
object StorageUtils {

  /**
   * Returns basic information of all RDDs persisted in the given SparkContext. This does not
   * include storage information.
   */
  def rddInfoFromSparkContext(sc: SparkContext): Array[RDDInfo] = {
    sc.persistentRdds.values.map { rdd =>
      val rddName = Option(rdd.name).getOrElse(rdd.id.toString)
      val rddNumPartitions = rdd.partitions.size
      val rddStorageLevel = rdd.getStorageLevel
      val rddInfo = new RDDInfo(rdd.id, rddName, rddNumPartitions, rddStorageLevel)
      rddInfo
    }.toArray
  }

  /** Returns storage information of all RDDs persisted in the given SparkContext. */
  def rddInfoFromStorageStatus(
      storageStatuses: Seq[StorageStatus],
      sc: SparkContext): Array[RDDInfo] = {
    rddInfoFromStorageStatus(storageStatuses, rddInfoFromSparkContext(sc))
  }

  /** Returns storage information of all RDDs in the given list. */
  def rddInfoFromStorageStatus(
      storageStatuses: Seq[StorageStatus],
      rddInfos: Seq[RDDInfo]): Array[RDDInfo] = {

    // Mapping from RDD ID -> an array of associated BlockStatuses
    val blockStatusMap = storageStatuses.flatMap(_.rddBlocks).toMap
      .groupBy { case (k, _) => k.rddId }
      .mapValues(_.values.toArray)

    // Mapping from RDD ID -> the associated RDDInfo (with potentially outdated storage information)
    val rddInfoMap = rddInfos.map { info => (info.id, info) }.toMap

    val rddStorageInfos = blockStatusMap.flatMap { case (rddId, blocks) =>
      // Add up memory, disk and Tachyon sizes
      val persistedBlocks =
        blocks.filter { status => status.memSize + status.diskSize + status.tachyonSize > 0 }
      val memSize = persistedBlocks.map(_.memSize).reduceOption(_ + _).getOrElse(0L)
      val diskSize = persistedBlocks.map(_.diskSize).reduceOption(_ + _).getOrElse(0L)
      val tachyonSize = persistedBlocks.map(_.tachyonSize).reduceOption(_ + _).getOrElse(0L)
      rddInfoMap.get(rddId).map { rddInfo =>
        rddInfo.numCachedPartitions = persistedBlocks.length
        rddInfo.memSize = memSize
        rddInfo.diskSize = diskSize
        rddInfo.tachyonSize = tachyonSize
        rddInfo
      }
    }.toArray

    scala.util.Sorting.quickSort(rddStorageInfos)
    rddStorageInfos
  }

  /** Returns a mapping from BlockId to the locations of the associated block. */
  def blockLocationsFromStorageStatus(
      storageStatuses: Seq[StorageStatus]): Map[BlockId, Seq[String]] = {
    val blockLocationPairs = storageStatuses.flatMap { storageStatus =>
      storageStatus.blocks.map { case (bid, _) => (bid, storageStatus.blockManagerId.hostPort) }
    }
    blockLocationPairs.toMap
      .groupBy { case (blockId, _) => blockId }
      .mapValues(_.values.toSeq)
  }

  /** Filters the given list of StorageStatus by the given RDD ID. */
  def filterStorageStatusByRDD(
      storageStatuses: Seq[StorageStatus],
      rddId: Int): Array[StorageStatus] = {
    storageStatuses.map { status =>
      val filteredBlocks = status.rddBlocks.filterKeys(_.rddId == rddId).toSeq
      val filteredBlockMap = mutable.Map[BlockId, BlockStatus](filteredBlocks: _*)
      new StorageStatus(status.blockManagerId, status.maxMem, filteredBlockMap)
    }.toArray
  }
}
